/*
 * FuncMeshCode.c
 *
 *  Created on : 2015.02.10
 *  Author     : BumpRecorder Co., Ltd.
 *  Web Page   : www.bumprecorder.com
 *  License    : CC BY-SA
 *
 */

#include <math.h>
#include "FuncMeshCode.h"


/*
 * return mesh code for latitude
 */
long getMeshCodeLat(double latitude, long size) {

	double	a = 469367.123429181;
	double	y;
	long	code = 0;

	if(latitude < 999) {
		y = latitude / 180 * 3.141592;
		if(0 <= y) {
			code =  (long)(log((1 + sin( y)) / (1 - sin( y))) / 2.0 * a);
		} else {
			code = -(long)(log((1 + sin(-y)) / (1 - sin(-y))) / 2.0 * a) - 1;
		}

		//	return mesh code for mesh size
		code = getMeshCodeSize(code, size);
	}

	return	code;
}

/*
 * return mesh code for longitude
 */
long getMeshCodeLon(double longitude, long size) {

	long	code = 0;

	if(longitude < 999) {
		if(0 <= longitude) {
			code =  (long)( 8192 * longitude);
		} else {
			code = -(long)(-8192 * longitude) - 1;
		}

		//	return mesh code for mesh size
		code = getMeshCodeSize(code, size);
	}

	return	code;
}

/*
 * return mesh code for mesh size
 */
long getMeshCodeSize(long code, long size) {

	long	s = size;
	long	codeSize, d;

	if(size <= 0) {
		s = 1;
	}

	if(0 <= code) {
		codeSize = code - code % s;
	} else {
		d = -code -1;
		codeSize = -(d - d % s) - s;
	}

	return	codeSize;
}
