/*
 * Sample code for "Grid Section" / "Square Mesh Section" calculations.
 * Created on : 2017.06.26
 * Author     : BumpRecorder Co., Ltd.
 * Web Page   : www.bumprecorder.com
 * License    : CC BY-SA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "FuncMeshCode.h"

#define		NUM_RING		64
#define		LENGTH_1DEG		(40000000 / 360)

typedef struct {
	long long int	time1st;
	long			size;
	long			iLat;
	long			iLon;
	long			pairLat;
	long			pairLon;
	long			next_iLat;
	long			next_iLon;
	double			lat1st;
	double			lon1st;
	double			lat;
	double			lon;
	double			sumLat;
	double			sumLon;
	double			L;
	double			num;
	double			sumSpeed;
	int				from_v;
	int				from_h;
	int				to_v;
	int				to_h;
	int				same;
}	RingData;

typedef struct {
	double			x;
	double			y;
}	Coordinate;

void	makeMeshFile	(char *in, char * out, int meshsize, int isMarge);
void	makeMeshData	(FILE *fpw, long long int sysTime, double lat, double lon, double speed, long size, int isMarge);
void	mergeMesh		(FILE *fpw);
void	flushMeshData	(FILE *fpw);
void	mergeData		(RingData *rd, int idx0, int idx1);
void	writeData		(FILE *fpw, long long int head);
double	getDistance2pos	(double lats, double lons, double late, double lone);
double	relationV2P		(Coordinate *base_from, Coordinate *base_to, Coordinate *point, double *pos100);
double	innerProduct	(Coordinate P, Coordinate Q);
double	norm			(Coordinate P);

static	RingData		gRing[NUM_RING];
static	long long int	gHeaderWrite;
static	long long int	gHeaderRead;


int main(int argc, const char * argv[]) {
    makeMeshFile("test_data.txt", "mesh.txt", 2, 1);
}

/*
 * The text file is converted to "Grid Section" File
 */
void makeMeshFile(char *in, char * out, int meshsize, int isMarge) {

	char			s[1024], oneLine[1024];
	FILE			*fpa, *fpw;
	long long int	devicetime;
	double			latitude, longitude, speed;

	if((fpa = fopen(in, "r")) != NULL) {
		fgets(oneLine, sizeof(oneLine), fpa);

		if((fpw = fopen(out, "w")) != NULL) {
			sprintf(s , "devicetime\tlatcode\tloncode\tlat1\tlon1\tlat2\tlon2\tl\tspeed\n");
			fputs(s, fpw);

			memset(&gRing, 0, sizeof(gRing));
			gHeaderWrite = 0;
			gHeaderRead  = 0;

			while(fgets(oneLine, sizeof(oneLine), fpa) != NULL) {
				sscanf(oneLine , "%lld%lf%lf%lf", &devicetime, &latitude, &longitude, &speed);
				makeMeshData(fpw, devicetime, latitude, longitude, speed, meshsize, isMarge);
			}
			flushMeshData(fpw);
			fclose(fpw);
		}
		fclose(fpa);
	}
}

/*
 * make grid data
 */
void makeMeshData(FILE *fpw, long long int devicetime, double lat, double lon, double speed, long size, int isMarge) {

    long			iLat, iLon;
    int				idx_c, idx_f;

	idx_c = gHeaderWrite % NUM_RING;

	iLat = getMeshCodeLat(lat, size);
	iLon = getMeshCodeLon(lon, size);

	if((gRing[idx_c].iLat != iLat) || (gRing[idx_c].iLon != iLon)) {
		if(0 < gRing[idx_c].num) {
			if((gRing[idx_c].lat1st < 999) && (gRing[idx_c].lat < 999)) {
				gRing[idx_c].next_iLat = iLat;
				gRing[idx_c].next_iLon = iLon;

				gRing[idx_c].L		= getDistance2pos(gRing[idx_c].lat1st, gRing[idx_c].lon1st, gRing[idx_c].lat, gRing[idx_c].lon);

				if(gRing[idx_c].iLat != iLat) {
					gRing[idx_c].to_v = 1;
				}
				if(gRing[idx_c].iLon != iLon) {
					gRing[idx_c].to_h = 1;
				}

				if((gRing[idx_c].from_v == gRing[idx_c].to_v) && (gRing[idx_c].from_h == gRing[idx_c].to_h)) {
					//	enter and exit direction are same
					gRing[idx_c].same = 1;
				} else {
					gRing[idx_c].same = 0;
				}

				idx_f = idx_c;
				gHeaderWrite = gHeaderWrite + 1;
				idx_c = gHeaderWrite % NUM_RING;
				gRing[idx_c].from_v = gRing[idx_f].to_v;
				gRing[idx_c].from_h = gRing[idx_f].to_h;
			}
		}

		if(isMarge) {
			//	try to merge
			mergeMesh(fpw);

		} else {
			//	not try to merge
			while(gHeaderRead < gHeaderWrite) {
				writeData(fpw, gHeaderRead);
				gHeaderRead = gHeaderRead + 1;
			}
		}

		gRing[idx_c].pairLat	= 0;
		gRing[idx_c].pairLon	= 0;
		gRing[idx_c].num		= 0;
		gRing[idx_c].time1st	= devicetime;
		gRing[idx_c].size		= size;
		gRing[idx_c].lat1st		= lat;
		gRing[idx_c].lon1st		= lon;
		gRing[idx_c].sumLat		= 0;
		gRing[idx_c].sumLon		= 0;
		gRing[idx_c].sumSpeed	= 0;
		gRing[idx_c].to_v		= 0;
		gRing[idx_c].to_h		= 0;
	}

	gRing[idx_c].sumSpeed	= gRing[idx_c].sumSpeed		+ speed;
	gRing[idx_c].num		= gRing[idx_c].num  + 1;

	gRing[idx_c].lat    = lat;
	gRing[idx_c].lon    = lon;
	gRing[idx_c].iLat   = iLat;
	gRing[idx_c].iLon   = iLon;
	gRing[idx_c].sumLat = gRing[idx_c].sumLat + lat;
	gRing[idx_c].sumLon = gRing[idx_c].sumLon + lon;
}

/*
 * merge mesh section
 */
void mergeMesh(FILE *fpw) {

    int		idx0, idx1, idx2;
    int		judge = 0;		//	1 = not merge    2 = merge with next
    long	size;

	while(gHeaderRead + 2 < gHeaderWrite) {
		idx0 =  gHeaderRead      % NUM_RING;
		idx1 = (gHeaderRead + 1) % NUM_RING;
		idx2 = (gHeaderRead + 2) % NUM_RING;

		size = gRing[idx0].size;
		if((labs(gRing[idx0].next_iLat -  gRing[idx0].iLat) <= size)
		&& (labs(gRing[idx0].next_iLon -  gRing[idx0].iLon) <= size)
		&& (     gRing[idx0].next_iLat == gRing[idx1].iLat         )
		&& (     gRing[idx0].next_iLon == gRing[idx1].iLon         )) {
			//	current and next section are continued
			if(gRing[idx0].same) {
				//	current is not required merge  [S]　vv or hh
				judge = 1;		//	current is not required merge

			} else {
				//	current is required merge with next  [d]　vh or hv
				if(gRing[idx1].same) {
					//	next is not required merge  [dS]　vv or hh
					judge = 1;		//	current is not required merge, because next is not required merge

				} else {
					//	next is required merge  [dd]　vh or hv
					if(gRing[idx2].same) {
						//	next next is not requred merge  [ddS]　vv or hh
						judge = 2;		//　merge current and next

					} else {
						if((labs(gRing[idx1].next_iLat -  gRing[idx1].iLat) <= size)
						&& (labs(gRing[idx1].next_iLon -  gRing[idx1].iLon) <= size)
						&& (     gRing[idx1].next_iLat == gRing[idx2].iLat         )
						&& (     gRing[idx1].next_iLon == gRing[idx2].iLon         )) {
							//	next and next next are continued
							//	next next is required merge  [ddd]
							if(gRing[idx2].L * 1.2 < gRing[idx0].L) {
								//	current length is more than 20% longer than next next length
								judge = 1;		//	current is not required merge

							} else if(gRing[idx0].L * 1.2 < gRing[idx2].L) {
								//	current length is less than 20% shorter than next next length
								judge = 2;		//　current is required merge

							} else {
								//	current and next next are almost same length
								if(gRing[idx0].iLat != gRing[idx1].iLat) {
									//	current and next are located north-south position
									judge = 2;		//　current is required merge
								} else {
									//	current and next are located east-west position
									judge = 1;		//	current is not required merge
								}
							}

						} else {
							//	next and next next are not continued  [dd-]
							judge = 2;		//　current is requried merge
						}
					}
				}
			}

		} else {
			//	current is not continued
			judge = 1;		//	current is not required merge
		}

		if(judge == 2) {
			//	current is required merge
			mergeData(gRing, idx0, idx1);
			gHeaderRead = gHeaderRead + 1;
		}

		writeData(fpw, gHeaderRead);
		gHeaderRead = gHeaderRead + 1;
	}
}

/*
 * flush mesh data
 */
void flushMeshData(FILE *fpw) {

    int		idx0, idx1;
    int		judge = 0;
    int		size = 2;

	while(gHeaderRead < gHeaderWrite) {
		switch(gHeaderWrite - gHeaderRead) {
		case 1:
			//	remaining is only current, it is not flush data because last section is mostly stopping in the section
			gHeaderRead = gHeaderRead + 1;
			break;

		case 2:
			//	remaining are current and next
			idx0 =  gHeaderRead      % NUM_RING;
			idx1 = (gHeaderRead + 1) % NUM_RING;

			if((labs(gRing[idx0].next_iLat -  gRing[idx0].iLat) <= size)
			&& (labs(gRing[idx0].next_iLon -  gRing[idx0].iLon) <= size)
			&& (     gRing[idx0].next_iLat == gRing[idx1].iLat         )
			&& (     gRing[idx0].next_iLon == gRing[idx1].iLon         )) {
				//	current and next are continued
				if(gRing[idx0].same) {
					//	current is not required merge  [S]　vv or hh
					judge = 1;		//	current is not required merge

				} else {
					//	current is required merge  [d]　vh or hv
					if(gRing[idx1].same) {
						//	next is not required merge  [dS]　vv or hh
						judge = 1;		//	current is not required merge, because next is not required merge

					} else {
						//	next is required merge  [dd]　vh or hv
						judge = 2;		//　current is required merge
					}
				}

			} else {
				//	current and next are not continued
				judge = 1;		//	current is not required merge
			}

			if(judge == 2) {
				//	merge current and next
				mergeData(gRing, idx0, idx1);
				gHeaderRead = gHeaderRead + 1;
			}

			writeData(fpw, gHeaderRead);
			gHeaderRead = gHeaderRead + 1;
			break;

		default :
			mergeMesh(fpw);
		}
	}
}

/*
 * data merge current section and next section
 */
void mergeData(RingData *rd, int idx0, int idx1) {

	if(rd[idx1].iLat < rd[idx0].iLat) {
		rd[idx1].pairLat = rd[idx0].iLat;
		rd[idx1].pairLon = rd[idx0].iLon;

	} else if(rd[idx1].iLat > rd[idx0].iLat) {
		rd[idx1].pairLat = rd[idx1].iLat;
		rd[idx1].pairLon = rd[idx1].iLon;
		rd[idx1].iLat    = rd[idx0].iLat;
		rd[idx1].iLon    = rd[idx0].iLon;

	} else if(rd[idx1].iLon < rd[idx0].iLon) {
		rd[idx1].pairLat = rd[idx0].iLat;
		rd[idx1].pairLon = rd[idx0].iLon;

	} else {
		rd[idx1].pairLat = rd[idx1].iLat;
		rd[idx1].pairLon = rd[idx1].iLon;
		rd[idx1].iLat    = rd[idx0].iLat;
		rd[idx1].iLon    = rd[idx0].iLon;
	}

	rd[idx1].time1st		= rd[idx0].time1st;
	rd[idx1].size			= rd[idx0].size;
	rd[idx1].lat1st			= rd[idx0].lat1st;
	rd[idx1].lon1st			= rd[idx0].lon1st;
	rd[idx1].sumLat			= rd[idx0].sumLat		+ rd[idx1].sumLat;
	rd[idx1].sumLon			= rd[idx0].sumLon		+ rd[idx1].sumLon;
	rd[idx1].L				= rd[idx0].L			+ rd[idx1].L;
	rd[idx1].num			= rd[idx0].num			+ rd[idx1].num;
	rd[idx1].sumSpeed		= rd[idx0].sumSpeed		+ rd[idx1].sumSpeed;
	rd[idx1].from_v			= rd[idx0].from_v;
	rd[idx1].from_h			= rd[idx0].from_h;

	if((rd[idx1].from_v == rd[idx1].to_v) && (rd[idx1].from_h == rd[idx1].to_h)) {
		//	enter and exit are same direction
		rd[idx1].same = 1;
	} else {
		rd[idx1].same = 0;
	}
}

/*
 * data output
 */
void writeData(FILE *fpw, long long int head) {

	char			s[1024] = {0};
    int				idx;
    Coordinate		Ps, Pe, Pc;				//	enter point, exit point, average point
    double			L, d, pos100, angle;

	idx = head % NUM_RING;

	if(0 < gRing[idx].num) {
		Ps.x = gRing[idx].lon1st;
		Ps.y = gRing[idx].lat1st;

		Pe.x = gRing[idx].lon;
		Pe.y = gRing[idx].lat;

		Pc.x = gRing[idx].sumLon / gRing[idx].num;
		Pc.y = gRing[idx].sumLat / gRing[idx].num;

		L = getDistance2pos(Ps.y, Ps.x, Pe.y, Pe.x);		//	distance between enter and exit point

		if(0 < L) {
			d = relationV2P(&Ps, &Pe, &Pc, &pos100);
			angle = 2 * atan(4 * d / L) * 180 / 3.141592;

			if(angle < 45) {
				sprintf(s, "%lld\t%ld\t%ld"  ,    gRing[idx].time1st, gRing[idx].iLat, gRing[idx].iLon);
				sprintf(s, "%s\t%.6f\t%.6f"  , s, gRing[idx].lat1st , gRing[idx].lon1st);
				sprintf(s, "%s\t%.6f\t%.6f"  , s, gRing[idx].lat    , gRing[idx].lon);
				sprintf(s, "%s\t%.1f\t%.2f\n", s, gRing[idx].L      , gRing[idx].sumSpeed / gRing[idx].num);
				fputs(s, fpw);
			}
		}
	}
}

/*
 * return distance [m] between 2 position
 */
double getDistance2pos(double lats, double lons, double late, double lone) {
	double	dLat, dLon, dL;

	dLat	= (late - lats) / 180.0 * 20000000.0;
	dLon	= (lone - lons) / 180.0 * 20000000.0 * cos(lats / 180 * 3.141592);
	dL		= sqrt(dLat * dLat + dLon * dLon);

	return	dL;
}

/*
 * distance and foot position for vector and point
 */
double relationV2P(Coordinate *base_from, Coordinate *base_to, Coordinate *point, double *pos100) {

	Coordinate	Pd, Qp;
	double		Lp, length_lon, normPd;
	double		distance;

	length_lon = cos(base_from->y / 180 * 3.141592) * LENGTH_1DEG;

	Pd.x = (base_to->x - base_from->x) * length_lon;
	Pd.y = (base_to->y - base_from->y) * LENGTH_1DEG;

	Qp.x = (point->x   - base_from->x) * length_lon;
	Qp.y = (point->y   - base_from->y) * LENGTH_1DEG;

	normPd = norm(Pd);

	Lp = innerProduct(Pd, Qp) / normPd;
	*pos100 = Lp / normPd;										//	foot position

	//	distance [m]
	distance = innerProduct(Qp, Qp) - Lp*Lp;
	if(0 < distance) {
		distance = sqrt(distance);
	} else {
		distance = 0;
	}

	return	distance;
}

/*
 * return inner product
 */
double innerProduct(Coordinate P, Coordinate Q) {
	return	P.x * Q.x + P.y * Q.y;
}

/*
 * return length of vector
 */
double norm(Coordinate P) {
	return	sqrt(innerProduct(P, P));
}
